/*
 * AutomatonToXMLComposer.h
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#ifndef AUTOMATON_TO_XML_COMPOSER_H_
#define AUTOMATON_TO_XML_COMPOSER_H_

#include <string>
#include <list>
#include "Automaton.h"
#include "FSM/EpsilonNFA.h"
#include "FSM/MultiInitialStateNFA.h"
#include "FSM/NFA.h"
#include "FSM/DFA.h"
#include "FSM/CompactNFA.h"
#include "FSM/ExtendedNFA.h"
#include "PDA/InputDrivenNPDA.h"
#include "PDA/VisiblyPushdownDPDA.h"
#include "PDA/VisiblyPushdownNPDA.h"
#include "PDA/RealTimeHeightDeterministicDPDA.h"
#include "PDA/RealTimeHeightDeterministicNPDA.h"
#include "PDA/NPDA.h"
#include "PDA/SinglePopNPDA.h"
#include "TM/OneTapeDTM.h"
#include "../sax/Token.h"

namespace alib {

template<typename T>
struct xmlApi;

} /* namespace alib */

namespace automaton {

/**
 * This class contains methods to print XML representation of automata to the output stream.
 */
class AutomatonToXMLComposer {
	void composeStates(std::list<sax::Token>&, const std::set<State>& states) const;
	void composeInputAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeCallInputAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeReturnInputAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeLocalInputAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeInitialStates(std::list<sax::Token>&, const std::set<State>& states) const;
	void composeInitialState(std::list<sax::Token>&, const State& state) const;
	void composeFinalStates(std::list<sax::Token>&, const std::set<State>& states) const;
	void composeStackAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeInitialStackSymbols(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeInitialStackSymbol(std::list<sax::Token>&, const alphabet::Symbol& symbols) const;
	void composeTapeAlphabet(std::list<sax::Token>&, const std::set<alphabet::Symbol>& symbols) const;
	void composeBlankSymbol(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeBottomOfTheStackSymbol(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeInputToPushdownStoreOperation(std::list<sax::Token>&, const automaton::InputDrivenNPDA& automaton) const;

	void composeTransitions(std::list<sax::Token>&, const EpsilonNFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const CompactNFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const ExtendedNFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const MultiInitialStateNFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const NFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const DFA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const DPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const SinglePopDPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const InputDrivenNPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const VisiblyPushdownDPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const VisiblyPushdownNPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const RealTimeHeightDeterministicDPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const RealTimeHeightDeterministicNPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const NPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const SinglePopNPDA& automaton) const;
	void composeTransitions(std::list<sax::Token>&, const OneTapeDTM& automaton) const;

	void composeTransitionTo(std::list<sax::Token>&, const State& state) const;
	void composeTransitionFrom(std::list<sax::Token>&, const State& state) const;
	void composeTransitionShift(std::list<sax::Token>&, const Shift shift) const;
	void composeTransitionPop(std::list<sax::Token>&, const std::vector<alphabet::Symbol>& symbols) const;
	void composeTransitionSinglePop(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeTransitionPush(std::list<sax::Token>&, const std::vector<alphabet::Symbol>& symbols) const;
	void composeTransitionSinglePush(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeTransitionInputSymbol(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeTransitionInputEpsilonSymbol(std::list<sax::Token>&, const std::variant<string::Epsilon, alphabet::Symbol>&) const;
	void composeTransitionOutputSymbol(std::list<sax::Token>&, const alphabet::Symbol& symbol) const;
	void composeTransitionOutputEpsilonSymbol(std::list<sax::Token>&, const std::variant<string::Epsilon, alphabet::Symbol>&) const;
	void composeTransitionInputString(std::list<sax::Token>&, const string::LinearString& string) const;
	void composeTransitionInputRegexp(std::list<sax::Token>&, const regexp::RegExp& regexp) const;

	/**
	 * Prints XML representation of Automaton to the output stream.
	 * @param automaton automaton to print
	 * @return list of xml tokens representing the automaton
	 */
	void compose(std::list<sax::Token>& output, const AutomatonBase& automaton) const;

	void compose(std::list<sax::Token>& output, const Automaton& automaton) const;
	
	void compose(std::list<sax::Token>& output, const DFA& automaton) const;
	void compose(std::list<sax::Token>& output, const NFA& automaton) const;
	void compose(std::list<sax::Token>& output, const MultiInitialStateNFA& automaton) const;
	void compose(std::list<sax::Token>& output, const EpsilonNFA& automaton) const;
	void compose(std::list<sax::Token>& output, const ExtendedNFA& automaton) const;
	void compose(std::list<sax::Token>& output, const CompactNFA& automaton) const;
	void compose(std::list<sax::Token>& output, const DPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const SinglePopDPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const InputDrivenNPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const VisiblyPushdownNPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const VisiblyPushdownDPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const RealTimeHeightDeterministicNPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const RealTimeHeightDeterministicDPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const NPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const SinglePopNPDA& automaton) const;
	void compose(std::list<sax::Token>& output, const OneTapeDTM& automaton) const;

	void compose(std::list<sax::Token>& output, const State& state) const;

	template<typename T> friend class alib::xmlApi;
};

} /* namespace automaton */

#endif /* AUTOMATON_TO_XML_COMPOSER_H_ */

