/*
 * FiniteAutomatonFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "FiniteAutomatonFromStringParser.h"

#include "../../exception/AlibException.h"

#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

#include "../../StringApi.hpp"

namespace automaton {

Automaton FiniteAutomatonFromStringParser::parseAutomaton(std::istream& input) const {
	return parseAutomaton(input, std::set<FEATURES>({FEATURES::EPSILON_NFA, FEATURES::MULTI_INITIAL_STATE_NFA, FEATURES::NFA, FEATURES::DFA}));
}

Automaton FiniteAutomatonFromStringParser::parseAutomaton(std::istream& input, const std::set<FEATURES>& features) const {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		if(!features.count(FEATURES::EPSILON_NFA)) throw exception::AlibException("Disabled formalism EpsilonNFA");
		m_FiniteAutomatonLexer.putback(input, token);
		return Automaton(parseEpsilonNFA(input));
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::MULTI_INITIAL_STATE_NFA) {
		if(!features.count(FEATURES::MULTI_INITIAL_STATE_NFA)) throw exception::AlibException("Disabled Formalism MultiInitialStateNFA");
		m_FiniteAutomatonLexer.putback(input, token);
		return Automaton(parseMultiInitialStateNFA(input));
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::NFA) {
		if(!features.count(FEATURES::NFA)) throw exception::AlibException("Disabled formalism NFA");
		m_FiniteAutomatonLexer.putback(input, token);
		return Automaton(parseNFA(input));
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::DFA) {
		if(!features.count(FEATURES::DFA)) throw exception::AlibException("Disabled formalism DFA");
		m_FiniteAutomatonLexer.putback(input, token);
		return Automaton(parseDFA(input));
	} else {
		throw exception::AlibException("Formalism not recognised (token = \"" + token.value + "\")");
	}
}

EpsilonNFA FiniteAutomatonFromStringParser::parseEpsilonNFA(std::istream& input) const {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		throw exception::AlibException("Unrecognised ENFA token.");
	}
	std::vector<std::variant<string::Epsilon, alphabet::Symbol> > symbols;

	token = m_FiniteAutomatonLexer.next(input);
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::ERROR) {
			m_FiniteAutomatonLexer.putback(input, token);
			alphabet::Symbol symbol = alib::stringApi<alphabet::Symbol>::parse(input);

			std::variant<string::Epsilon, alphabet::Symbol> symbolVariant(symbol);
			symbols.push_back(symbolVariant);
		} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON) {
			symbols.push_back(std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON));
		}

		token = m_FiniteAutomatonLexer.next(input);
	}

	State* initialState = NULL;
	std::set<State> finalStates;
	std::set<State> states;
	std::set<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, State>> transitionFunction;

	parseEpsilonNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = m_FiniteAutomatonLexer.next(input);

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = m_FiniteAutomatonLexer.next(input);
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			m_FiniteAutomatonLexer.putback(input, token);

		parseEpsilonNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = m_FiniteAutomatonLexer.next(input);
	}

	if(initialState == NULL) throw exception::AlibException("No initial state recognised.");

	EpsilonNFA res(*initialState);
	delete initialState;

	for(const std::variant<string::Epsilon, alphabet::Symbol> inputSymbol : symbols) {
		if(inputSymbol.is<alphabet::Symbol>())
			res.addInputSymbol(inputSymbol.get<alphabet::Symbol>());
	}
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, State> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return  res;
}

MultiInitialStateNFA FiniteAutomatonFromStringParser::parseMultiInitialStateNFA(std::istream& input) const {
	MultiInitialStateNFA res;

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::MULTI_INITIAL_STATE_NFA) {
		throw exception::AlibException("Unrecognised MISNFA token.");
	}
	std::vector<alphabet::Symbol> symbols;

	token = m_FiniteAutomatonLexer.next(input);
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		m_FiniteAutomatonLexer.putback(input, token);
		alphabet::Symbol symbol = alib::stringApi<alphabet::Symbol>::parse(input);
		res.addInputSymbol(symbol);
		symbols.push_back(symbol);

		token = m_FiniteAutomatonLexer.next(input);
	}

	parseMultiInitialStateNFATransition(input, res, symbols);
	token = m_FiniteAutomatonLexer.next(input);

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = m_FiniteAutomatonLexer.next(input);
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			m_FiniteAutomatonLexer.putback(input, token);

		parseMultiInitialStateNFATransition(input, res, symbols);
		token = m_FiniteAutomatonLexer.next(input);
	}

	return  res;
}

NFA FiniteAutomatonFromStringParser::parseNFA(std::istream& input) const {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::NFA) {
		throw exception::AlibException("Unrecognised NFA token.");
	}
	std::vector<alphabet::Symbol> symbols;

	token = m_FiniteAutomatonLexer.next(input);
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		m_FiniteAutomatonLexer.putback(input, token);
		alphabet::Symbol symbol = alib::stringApi<alphabet::Symbol>::parse(input);
		symbols.push_back(symbol);

		token = m_FiniteAutomatonLexer.next(input);
	}

	State* initialState = NULL;
	std::set<State> finalStates;
	std::set<State> states;
	std::set<std::tuple<State, alphabet::Symbol, State>> transitionFunction;

	parseNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = m_FiniteAutomatonLexer.next(input);

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = m_FiniteAutomatonLexer.next(input);
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			m_FiniteAutomatonLexer.putback(input, token);

		parseNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = m_FiniteAutomatonLexer.next(input);
	}

	if(initialState == NULL) throw exception::AlibException("No initial state recognised.");

	NFA res(*initialState);
	delete initialState;

	res.setInputAlphabet(std::set<alphabet::Symbol>(symbols.begin(), symbols.end()));
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const std::tuple<State, alphabet::Symbol, State> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return  res;
}

DFA FiniteAutomatonFromStringParser::parseDFA(std::istream& input) const {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::DFA) {
		throw exception::AlibException("Unrecognised DFA token.");
	}
	std::vector<alphabet::Symbol> symbols;

	token = m_FiniteAutomatonLexer.next(input);
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		m_FiniteAutomatonLexer.putback(input, token);
		alphabet::Symbol symbol = alib::stringApi<alphabet::Symbol>::parse(input);
		symbols.push_back(symbol);

		token = m_FiniteAutomatonLexer.next(input);
	}

	State* initialState = NULL;
	std::set<State> finalStates;
	std::set<State> states;
	std::set<std::tuple<State, alphabet::Symbol, State>> transitionFunction;

	parseDFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = m_FiniteAutomatonLexer.next(input);

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = m_FiniteAutomatonLexer.next(input);
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			m_FiniteAutomatonLexer.putback(input, token);

		parseDFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = m_FiniteAutomatonLexer.next(input);
	}

	if(initialState == NULL) throw exception::AlibException("No initial state recognised.");

	DFA res(*initialState);
	delete initialState;

	res.setInputAlphabet(std::set<alphabet::Symbol>(symbols.begin(), symbols.end()));
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const std::tuple<State, alphabet::Symbol, State> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return res;
}

void FiniteAutomatonFromStringParser::initialFinalState(std::istream& input, bool& initial, bool& final) const {
	initial = false;
	final = false;

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::IN) {
		initial = true;
		token = m_FiniteAutomatonLexer.next(input);

		if(token.type == FiniteAutomatonFromStringLexer::TokenType::OUT) {
			final = true;
		} else {
			m_FiniteAutomatonLexer.putback(input, token);
		}
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::OUT) {
		final = true;
		token = m_FiniteAutomatonLexer.next(input);

		if(token.type == FiniteAutomatonFromStringLexer::TokenType::IN) {
			initial = true;
		} else {
			m_FiniteAutomatonLexer.putback(input, token);
		}
	} else {
		m_FiniteAutomatonLexer.putback(input, token);
	}
}


void FiniteAutomatonFromStringParser::parseEpsilonNFATransition(std::istream& input, std::set<State>& states, const std::vector<std::variant<string::Epsilon, alphabet::Symbol>>& symbols, State*& initialState, std::set<State>& finalStates, std::set<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, State>>& transitionFunction) const {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	State from(alib::stringApi<label::Label>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::AlibException("Multiple initial states are not avaiable for NFA type");
		initialState = new State(from);
	}
	if(final) finalStates.insert(from);

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	std::vector<std::variant<string::Epsilon, alphabet::Symbol>>::const_iterator iter = symbols.begin();

	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::AlibException("Invalid line format");

		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
			m_FiniteAutomatonLexer.putback(input, token);
			do {
				State to(alib::stringApi<label::Label>::parse(input));
				states.insert(to);
				transitionFunction.insert(std::make_tuple(from, *iter, to));

				token = m_FiniteAutomatonLexer.next(input);
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = m_FiniteAutomatonLexer.next(input);
		}
		iter++;
	}
	m_FiniteAutomatonLexer.putback(input, token);

	if(iter != symbols.end()) throw exception::AlibException("Invalid line format");
}

void FiniteAutomatonFromStringParser::parseMultiInitialStateNFATransition(std::istream& input, MultiInitialStateNFA& res, const std::vector<alphabet::Symbol>& symbols) const {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	State from(alib::stringApi<label::Label>::parse(input));
	res.addState(from);
	if(initial) res.addInitialState(from);
	if(final) res.addFinalState(from);

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	std::vector<alphabet::Symbol>::const_iterator iter = symbols.begin();

	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::AlibException("Invalid line format");

		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
			m_FiniteAutomatonLexer.putback(input, token);
			do {
				State to(alib::stringApi<label::Label>::parse(input));
				res.addState(to);
				res.addTransition(from, *iter, to);

				token = m_FiniteAutomatonLexer.next(input);
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = m_FiniteAutomatonLexer.next(input);
		}
		iter++;
	}
	m_FiniteAutomatonLexer.putback(input, token);

	if(iter != symbols.end()) throw exception::AlibException("Invalid line format");
}

void FiniteAutomatonFromStringParser::parseNFATransition(std::istream& input, std::set<State>& states, const std::vector<alphabet::Symbol>& symbols, State*& initialState, std::set<State>& finalStates, std::set<std::tuple<State, alphabet::Symbol, State>>& transitionFunction) const {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	State from(alib::stringApi<label::Label>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::AlibException("Multiple initial states are not avaiable for NFA type");
		initialState = new State(from);
	}
	if(final) finalStates.insert(from);

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	std::vector<alphabet::Symbol>::const_iterator iter = symbols.begin();

	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::AlibException("Invalid line format");

		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
			m_FiniteAutomatonLexer.putback(input, token);
			do {
				State to(alib::stringApi<label::Label>::parse(input));
				states.insert(to);
				transitionFunction.insert(std::make_tuple(from, *iter, to));

				token = m_FiniteAutomatonLexer.next(input);
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = m_FiniteAutomatonLexer.next(input);
		}
		iter++;
	}
	m_FiniteAutomatonLexer.putback(input, token);

	if(iter != symbols.end()) throw exception::AlibException("Invalid line format");
}

void FiniteAutomatonFromStringParser::parseDFATransition(std::istream& input, std::set<State>& states, const std::vector<alphabet::Symbol>& symbols, State*& initialState, std::set<State>& finalStates, std::set<std::tuple<State, alphabet::Symbol, State>>& transitionFunction) const {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	State from(alib::stringApi<label::Label>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::AlibException("Multiple initial states are not avaiable for NFA type");
		initialState = new State(from);
	}
	if(final) finalStates.insert(from);

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next(input);
	std::vector<alphabet::Symbol>::const_iterator iter = symbols.begin();

	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::AlibException("Invalid line format");

		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
			m_FiniteAutomatonLexer.putback(input, token);
			State to(alib::stringApi<label::Label>::parse(input));
			states.insert(to);
			transitionFunction.insert(std::make_tuple(from, *iter, to));

			token = m_FiniteAutomatonLexer.next(input);
		} else {
			token = m_FiniteAutomatonLexer.next(input);
		}
		iter++;
	}
	m_FiniteAutomatonLexer.putback(input, token);

	if(iter != symbols.end()) throw exception::AlibException("Invalid line format");
}

} /* namespace automaton */
