/*
 * VisiblyPushdownDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "VisiblyPushdownDPDA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

namespace automaton {

VisiblyPushdownDPDA::VisiblyPushdownDPDA(const State& initialState, const alphabet::Symbol& bottomOfTheStackSymbol) : SingleInitialState(initialState), BottomOfTheStackSymbolPushdownStoreAlphabet(bottomOfTheStackSymbol) {

}

AutomatonBase* VisiblyPushdownDPDA::clone() const {
	return new VisiblyPushdownDPDA(*this);
}

AutomatonBase* VisiblyPushdownDPDA::plunder() && {
	return new VisiblyPushdownDPDA(std::move(*this));
}

bool VisiblyPushdownDPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (state == callTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(callTransition.second.first == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (state == std::get<0>(returnTransition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(returnTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& localTransition : localTransitions) {
		if (state == localTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(localTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool VisiblyPushdownDPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (symbol == callTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (symbol == std::get<1>(returnTransition.first))
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& localTransition : localTransitions) {
		if (symbol == localTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return callInputAlphabet.erase(symbol) || returnInputAlphabet.erase(symbol) || localInputAlphabet.erase(symbol);
}

bool VisiblyPushdownDPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (symbol == callTransition.second.second)
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (symbol == std::get<2>(returnTransition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(bottomOfTheStackSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool VisiblyPushdownDPDA::addCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (callInputAlphabet.find(input) == callInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(push) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);

	if(callTransitions.find(key) != callTransitions.end() && callTransitions.find(key)->second == value)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == from && std::get<1>(transition.first) == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	callTransitions.insert(std::make_pair(key, value));
	return true;
}

bool VisiblyPushdownDPDA::addReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (returnInputAlphabet.find(input) == returnInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(from, input, pop);

	if(returnTransitions.find(key) != returnTransitions.end() && returnTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == from && std::get<1>(transition.first) == input && std::get<2>(transition.first) == pop)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	returnTransitions.insert(std::make_pair(key, to));
	return true;
}

bool VisiblyPushdownDPDA::addLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (localInputAlphabet.find(input) == localInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(from, input);
	
	if(localTransitions.find(key) != localTransitions.end() && localTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == from && std::get<1>(transition.first) == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == from && transition.first.second == input)
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	localTransitions.insert(std::make_pair(key, to));
	return true;
}

bool VisiblyPushdownDPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, alphabet::Symbol> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);
	
	if (callTransitions.find(key) == callTransitions.end())
		return false;

	if(callTransitions.find(key)->second != value)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	callTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(from, input, pop);

	if (returnTransitions.find(key) == returnTransitions.end())
		return false;

	if(returnTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	returnTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key(from, input);

	if (localTransitions.find(key) == localTransitions.end())
		return false;

	if(localTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	localTransitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& VisiblyPushdownDPDA::getCallTransitions() const {
		return callTransitions;
}

const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getReturnTransitions() const {
		return returnTransitions;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getLocalTransitions() const {
		return localTransitions;
}

int VisiblyPushdownDPDA::compare(const VisiblyPushdownDPDA& other) const {
	auto first = std::tie(states, callInputAlphabet, returnInputAlphabet, localInputAlphabet, initialState, finalStates, stackAlphabet, bottomOfTheStackSymbol, callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.states, other.callInputAlphabet, other.returnInputAlphabet, other.localInputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.bottomOfTheStackSymbol, other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void VisiblyPushdownDPDA::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownDPDA "
		<< "states = " << states
		<< "callInputAlphabet = " << callInputAlphabet
		<< "returnInputAlphabet = " << returnInputAlphabet
		<< "localInputAlphabet = " << localInputAlphabet
		<< "initialState = " << initialState
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "bottomOfTheStackSymbol = " << bottomOfTheStackSymbol
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

VisiblyPushdownDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */
