/*
 * VISIBLY_PUSHDOWN_DPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISIBLY_PUSHDOWN_DPDA_H_
#define VISIBLY_PUSHDOWN_DPDA_H_

#include <map>
#include <vector>
#include "../../std/map.hpp"
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/CallReturnLocalInputAlphabet.h"
#include "../common/BottomOfTheStackSymbolPushdownStoreAlphabet.h"
#include "../../alphabet/Symbol.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class VisiblyPushdownDPDA : public std::acceptor<VisiblyPushdownDPDA, VisitableAutomatonBase, std::acceptor<VisiblyPushdownDPDA, alib::VisitableObjectBase, AutomatonBase> >, public CallReturnLocalInputAlphabet, public SingleInitialState, public BottomOfTheStackSymbolPushdownStoreAlphabet {
protected:
	std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> > callTransitions;
	std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State > returnTransitions;
	std::map<std::pair<State, alphabet::Symbol>, State > localTransitions;
public:
	explicit VisiblyPushdownDPDA(const State& initialSymbol, const alphabet::Symbol& bottomOfTheStackSymbol);

	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/*
	 * @copydoc Automaton::removeStackSymbol(const Symbol&)
	 */
	virtual bool removeStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition(const State& current, const alphabet::Symbol& input, const State& next, const alphabet::Symbol& push);

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition(const State& current, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& next);

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition(const State& current, const alphabet::Symbol& input, const State& next);

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition(const State& current, const alphabet::Symbol& input, const State& next, const alphabet::Symbol& push);

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition(const State& current, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& next);

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition(const State& current, const alphabet::Symbol& input, const State& next);

	const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& getCallTransitions() const;

	const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& getReturnTransitions() const;

	const std::map<std::pair<State, alphabet::Symbol>, State>& getLocalTransitions() const;

	virtual int compare(const ObjectBase& other) const {
		return -other.compare(*this);
	}

	virtual int compare(const VisiblyPushdownDPDA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual operator std::string() const;

	virtual int selfTypeId() const {
		return typeId<VisiblyPushdownDPDA>();
	}
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::VisiblyPushdownDPDA> {
	int operator()(const automaton::VisiblyPushdownDPDA& first, const automaton::VisiblyPushdownDPDA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* VISIBLY_PUSHDOWN_DPDA_H_ */
