/*
 * ContainerFromXMLParser.h
 *
 *  Created on: Oct 12, 2013
 *      Author: Jan Travnicek
 */

#ifndef CONTAINER_FROM_XML_PARSER_H_
#define CONTAINER_FROM_XML_PARSER_H_

#include "../std/variant.hpp"
#include "../sax/FromXMLParserHelper.h"
#include "ContainerFeatures.h"

#include "../object/Object.h"
#include "ObjectsSet.h"
#include "ObjectsVector.h"
#include "ObjectsPair.h"
#include "ObjectsMap.h"

#include "Container.h"
#include "../exception/AlibException.h"

namespace alib {

template<typename T>
struct xmlApi;

} /* namespace alib */

namespace container {

/**
 * Parser used to get general FSM or EpsilonNFA, NFA, DFA from XML parsed into list of Tokens.
 */
class ContainerFromXMLParser : public sax::FromXMLParserHelper {
	Container parseContainer(std::list<sax::Token>& input) const;
	Container parseContainer(std::list<sax::Token>& input, const std::set<FEATURES>& features) const;

	template<typename T>
	std::set<T> parseSet(std::list<sax::Token>& input) const;

	ObjectsSet parseObjectsSet(std::list<sax::Token>& input) const;

	template<typename T>
	std::vector<T> parseVector(std::list<sax::Token>& input) const;

	ObjectsVector parseObjectsVector(std::list<sax::Token>& input) const;

	template<typename T, typename R>
	std::pair<T, R> parsePair(std::list<sax::Token>& input) const;

	ObjectsPair parseObjectsPair(std::list<sax::Token>& input) const;

	template<typename T, typename R>
	std::map<T, R> parseMap(std::list<sax::Token>& input) const;

	ObjectsMap parseObjectsMap(std::list<sax::Token>& input) const;

	template<typename ... Ts>
	std::variant<Ts...> parseVariant(std::list<sax::Token>& input) const;

	template<typename T> friend class alib::xmlApi;

public:
	bool first(const std::list<sax::Token>& input) const;
};

} /* namespace container */

#include "../XmlApi.hpp"

namespace container {

template<typename T>
std::set<T> ContainerFromXMLParser::parseSet(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::CONTAINER_OBJECTS_SET);

	std::set<T> set;

	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		set.insert(alib::xmlApi<T>::parse(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::CONTAINER_OBJECTS_SET);
	return set;
}

template<typename T>
std::vector<T> ContainerFromXMLParser::parseVector(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::CONTAINER_OBJECTS_VECTOR);

	std::vector<T> vector;

	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		vector.push_back(alib::xmlApi<T>::parse(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::CONTAINER_OBJECTS_VECTOR);
	return vector;
}

template<typename T, typename R>
std::pair<T, R> ContainerFromXMLParser::parsePair(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::CONTAINER_OBJECTS_PAIR);

	T first = alib::xmlApi<T>::parse(input);
	R second = alib::xmlApi<R>::parse(input);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::CONTAINER_OBJECTS_PAIR);
	return std::make_pair(first, second);
}

template<typename T, typename R>
std::map<T, R> ContainerFromXMLParser::parseMap(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::CONTAINER_OBJECTS_MAP);

	std::map<T, R> map;

	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		map.insert(alib::xmlApi<std::pair<T, R>>::parse(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::CONTAINER_OBJECTS_MAP);
	return map;
}

/* TODO for gcc-4.9 uncoment this
template<typename T>
T parseVariantHelper(std::list<sax::Token>&) {
	throw exception::AlibException("Can't parse variant type");
}

template<typename T, typename R, typename ... Ts>
T parseVariantHelper(std::list<sax::Token>& input) {
	if(alib::xmlApi<R>::first(input)) return T(alib::xmlApi<R>::parse(input));
	return parseVariantHelper<T, Ts...>(input);
}

template<typename ... Ts>
std::variant<Ts...> ContainerFromXMLParser::parseVariant(std::list<sax::Token>& input) const {
	return parseVariantHelper<std::variant<Ts...>, Ts...>(input);
}
*/

} /* namespace container */

#endif /* CONTAINER_FROM_XML_PARSER_H_ */

