/*
 * XmlDataFactory.hpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Jan Travnicek
 */

#ifndef DATA_FACTORY_H_
#define DATA_FACTORY_H_

#include <string>
#include <list>

#include "../sax/Token.h"
#include "../sax/SaxParseInterface.h"
#include "../sax/SaxComposeInterface.h"
#include "../XmlApi.hpp"

namespace alib {

/**
 * Data builder.
 */
class XmlDataFactory {
public:
	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template<class T>
	static T fromFile(const std::string& filename) {
		std::list<sax::Token> tokens;
		sax::SaxParseInterface::parseFile(filename, tokens);
		return fromTokens<T>(tokens);
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template<class T>
	static T fromString(const std::string& str) {
		std::list<sax::Token> tokens;
		sax::SaxParseInterface::parseMemory(str, tokens);
		return fromTokens<T>(tokens);
	}
	
	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template<class T>
	static T fromStdin() {
		return fromStream<T>(std::cin);
	}
	
	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template<class T>
	static T fromStream(std::istream& in) {
		std::string input{std::istreambuf_iterator<char>(in), std::istreambuf_iterator<char>()};
		return fromString<T>(input);
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template<class T>
	static T fromTokens(std::list<sax::Token>& tokens) {
		if(tokens.size() == 0) throw exception::AlibException("Empty tokens list");

		if(alib::xmlApi<exception::AlibException>::first(tokens)) throw alib::xmlApi<exception::AlibException>::parse(tokens);

		T res = alib::xmlApi<T>::parse(tokens);
		if(tokens.size() != 0) throw exception::AlibException("Unexpeted tokens at the end of the xml");
		return res;
	}

	/**
	 * Determines whether the input token stream contains given type
	 * @param tokens
	 * @return bool
	 */
	template<class T>
	static bool first(const std::list<sax::Token>& tokens) {
		return alib::xmlApi<exception::AlibException>::first(tokens) || alib::xmlApi<T>::first(tokens);
	}

	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template<class T>
	static void toFile(const T& data, const std::string& filename) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		sax::SaxComposeInterface::printFile(filename, tokens);
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template<class T>
	static std::string toString(const T& data) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		std::string str;
		sax::SaxComposeInterface::printMemory(str, tokens);
		return str;
	}
	
	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template<class T>
	static void toStdout(const T& data) {
		toStream<T>(data, std::cout);
	}
	
	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template<class T>
	static void toStream(const T& data, std::ostream& out) {
		std::list<sax::Token> tokens = toTokens<T>(data);
		sax::SaxComposeInterface::printStream(out, tokens);
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template<class T>
	static std::list<sax::Token> toTokens(const T& data) {
		std::list<sax::Token> res;
		alib::xmlApi<T>::compose(res, data);
		return res;
	}

};

template<>
inline bool XmlDataFactory::first<exception::AlibException>(const std::list<sax::Token>& tokens) {
	return alib::xmlApi<exception::AlibException>::first(tokens);
}

template<>
inline exception::AlibException XmlDataFactory::fromTokens<exception::AlibException>(std::list<sax::Token>& tokens) {
	if(tokens.size() == 0) throw exception::AlibException("Empty tokens list");

	exception::AlibException res = alib::xmlApi<exception::AlibException>::parse(tokens);
	if(tokens.size() != 0) throw exception::AlibException("Unexpeted tokens at the end of the xml");
	return res;
}

} /* namespace alib */

#endif /* DATA_FACTORY_H_ */
