/*
 * LabelFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LabelFromStringParser.h"
#include "../exception/AlibException.h"
#include "PrimitiveLabel.h"
#include "ObjectLabel.h"
#include "LabelSetLabel.h"
#include "LabelPairLabel.h"
#include <algorithm>

#include "../StringApi.hpp"

namespace label {

Label LabelFromStringParser::parseLabel(std::istream& input) const {
	return parseLabel(input, std::set<FEATURES>({FEATURES::PRIMITIVE, FEATURES::HEXAVIGESIMAL, FEATURES::OBJECT, FEATURES::LABEL_SET, FEATURES::LABEL_PAIR, FEATURES::UNIQUE_LABEL}));
}

Label LabelFromStringParser::parseLabel(std::istream& input, const std::set<FEATURES>& features) const {
	LabelFromStringLexer::Token token = m_Lexer.next(input);
	switch(token.type) {
	case LabelFromStringLexer::TokenType::TEOF:
		throw exception::AlibException("Unexpected end of file");
	case LabelFromStringLexer::TokenType::SET_BEGIN: {
		if(!features.count(FEATURES::LABEL_SET)) throw exception::AlibException();
		token = m_Lexer.next(input);

		std::set<Label> labelSet;
		if(token.type != LabelFromStringLexer::TokenType::SET_END) while(true) {
			m_Lexer.putback(input, std::move(token));
			Label innerLabel = this->parseLabel(input);
			labelSet.insert(innerLabel);

			token = m_Lexer.next(input);
			if(token.type != LabelFromStringLexer::TokenType::COMMA) 
				break;

			token = m_Lexer.next(input);
		}

		if(token.type != LabelFromStringLexer::TokenType::SET_END)
			throw exception::AlibException("Expected SET_END token.");
		return Label(label::LabelSetLabel(labelSet));
	}
	case LabelFromStringLexer::TokenType::PAIR_BEGIN: {
		if(!features.count(FEATURES::LABEL_PAIR)) throw exception::AlibException();

		Label firstLabel = this->parseLabel(input);

		token = m_Lexer.next(input);
		if(token.type != LabelFromStringLexer::TokenType::COMMA)
			throw exception::AlibException("Excepted COMMA token.");

		Label secondLabel = this->parseLabel(input);

		token = m_Lexer.next(input);
		if(token.type != LabelFromStringLexer::TokenType::PAIR_END)
			throw exception::AlibException("Expected PAIR_END token.");

		return Label(label::LabelPairLabel(std::make_pair(firstLabel, secondLabel)));
	}
	case LabelFromStringLexer::TokenType::SET_END:
	case LabelFromStringLexer::TokenType::PAIR_END:
	case LabelFromStringLexer::TokenType::COMMA:
		throw exception::AlibException("Unexpected start of Label.");
	case LabelFromStringLexer::TokenType::ERROR:
		if(!features.count(FEATURES::PRIMITIVE)) throw exception::AlibException();
		m_Lexer.putback(input, token);
		return Label(PrimitiveLabel(alib::stringApi<primitive::Primitive>::parse(input)));
	}
	throw exception::AlibException("Unrecognised input \"" + token.value + "\"");
}

} /* namespace label */
