/*
 * RegExpFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpFromXMLParser.h"
#include "../sax/ParserException.h"

#include "../XmlApi.hpp"

namespace regexp {

RegExp RegExpFromXMLParser::parseRegExp(std::list<sax::Token>& input) const {
	return parseRegExp(input, std::set<FEATURES>({FEATURES::FORMAL, FEATURES::UNBOUNDED}));
}

RegExp RegExpFromXMLParser::parseRegExp(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(alib::xmlApi<UnboundedRegExp>::first(input)) {
		if(!features.count(FEATURES::UNBOUNDED)) throw exception::AlibException();
		return RegExp(parseUnboundedRegExp(input));
	} else if(alib::xmlApi<FormalRegExp>::first(input)) {
		if(!features.count(FEATURES::FORMAL)) throw exception::AlibException();
		return RegExp(parseFormalRegExp(input));
	}
	throw exception::AlibException();
}

bool RegExpFromXMLParser::first(const std::list<sax::Token>& input) const {
	if(alib::xmlApi<UnboundedRegExp>::first(input) || alib::xmlApi<FormalRegExp>::first(input)) {
		return true;
	} else {
		return false;
	}
}

UnboundedRegExp RegExpFromXMLParser::parseUnboundedRegExp(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::REGEXP_UNBOUNDED_REGEXP);

	UnboundedRegExp regexp;
	parseAlphabet(input, regexp);

	UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
	regexp.setRegExp(std::move(*element));
	delete element;

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::REGEXP_UNBOUNDED_REGEXP);
	return regexp;
}

FormalRegExp RegExpFromXMLParser::parseFormalRegExp(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::REGEXP_FORMAL_REGEXP);

	FormalRegExp regexp;
	parseAlphabet(input, regexp);

	FormalRegExpElement* element = parseFormalRegExpElement(input);
	regexp.setRegExp(std::move(*element));
	delete element;

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::REGEXP_FORMAL_REGEXP);
	return regexp;
}

void RegExpFromXMLParser::parseAlphabet(std::list<sax::Token> &input, UnboundedRegExp& regexp) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alphabet");
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		regexp.addSymbolToAlphabet(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
	popToken(input, sax::Token::TokenType::END_ELEMENT, "alphabet");
}

void RegExpFromXMLParser::parseAlphabet(std::list<sax::Token> &input, FormalRegExp& regexp) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alphabet");
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		regexp.addSymbolToAlphabet(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
	popToken(input, sax::Token::TokenType::END_ELEMENT, "alphabet");
}

UnboundedRegExpElement* RegExpFromXMLParser::parseUnboundedRegExpElement(std::list<sax::Token>& input) const {
	if (isToken(input, sax::Token::TokenType::START_ELEMENT, "empty")) {
		return parseUnboundedRegExpEmpty(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		return parseUnboundedRegExpEpsilon(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "iteration")) {
		return parseUnboundedRegExpIteration(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "alternation")) {
		return parseUnboundedRegExpAlternation(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation")) {
		return parseUnboundedRegExpConcatenation(input);
	} else {
		return new UnboundedRegExpSymbol(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
}

UnboundedRegExpAlternation* RegExpFromXMLParser::parseUnboundedRegExpAlternation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alternation");

	UnboundedRegExpAlternation* alternation = new UnboundedRegExpAlternation();

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
		alternation->appendElement(std::move(*element));
		delete element;
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "alternation");
	return alternation;
}

UnboundedRegExpConcatenation* RegExpFromXMLParser::parseUnboundedRegExpConcatenation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation");

	UnboundedRegExpConcatenation* concatenation = new UnboundedRegExpConcatenation();

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
		concatenation->appendElement(std::move(*element));
		delete element;
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "concatenation");
	return concatenation;

}

UnboundedRegExpIteration* RegExpFromXMLParser::parseUnboundedRegExpIteration(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "iteration");

	UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
	UnboundedRegExpIteration* iteration = new UnboundedRegExpIteration(std::move(*element));
	delete element;

	popToken(input, sax::Token::TokenType::END_ELEMENT, "iteration");
	return iteration;
}

UnboundedRegExpEpsilon* RegExpFromXMLParser::parseUnboundedRegExpEpsilon(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon");

	UnboundedRegExpEpsilon* epsilon = new UnboundedRegExpEpsilon();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	return epsilon;
}

UnboundedRegExpEmpty* RegExpFromXMLParser::parseUnboundedRegExpEmpty(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "empty");

	UnboundedRegExpEmpty* empty = new UnboundedRegExpEmpty();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "empty");
	return empty;
}

FormalRegExpElement* RegExpFromXMLParser::parseFormalRegExpElement(std::list<sax::Token>& input) const {
	if (isToken(input, sax::Token::TokenType::START_ELEMENT, "empty")) {
		return parseFormalRegExpEmpty(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		return parseFormalRegExpEpsilon(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "iteration")) {
		return parseFormalRegExpIteration(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "alternation")) {
		return parseFormalRegExpAlternation(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation")) {
		return parseFormalRegExpConcatenation(input);
	} else {
		return new FormalRegExpSymbol(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
}

FormalRegExpAlternation* RegExpFromXMLParser::parseFormalRegExpAlternation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alternation");

	FormalRegExpElement* element1 = parseFormalRegExpElement(input);
	FormalRegExpElement* element2 = parseFormalRegExpElement(input);
	
	FormalRegExpAlternation* alternation = new FormalRegExpAlternation(std::move(*element1), std::move(*element2));

	delete element1;
	delete element2;

	popToken(input, sax::Token::TokenType::END_ELEMENT, "alternation");
	return alternation;
}

FormalRegExpConcatenation* RegExpFromXMLParser::parseFormalRegExpConcatenation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation");

	FormalRegExpElement* element1 = parseFormalRegExpElement(input);
	FormalRegExpElement* element2 = parseFormalRegExpElement(input);

	FormalRegExpConcatenation* concatenation = new FormalRegExpConcatenation(std::move(*element1), std::move(*element2));

	delete element1;
	delete element2;

	popToken(input, sax::Token::TokenType::END_ELEMENT, "concatenation");
	return concatenation;

}

FormalRegExpIteration* RegExpFromXMLParser::parseFormalRegExpIteration(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "iteration");

	FormalRegExpElement* element = parseFormalRegExpElement(input);
	FormalRegExpIteration* iteration = new FormalRegExpIteration(std::move(*element));
	delete element;

	popToken(input, sax::Token::TokenType::END_ELEMENT, "iteration");
	return iteration;
}

FormalRegExpEpsilon* RegExpFromXMLParser::parseFormalRegExpEpsilon(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon");

	FormalRegExpEpsilon* epsilon = new FormalRegExpEpsilon();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	return epsilon;
}

FormalRegExpEmpty* RegExpFromXMLParser::parseFormalRegExpEmpty(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "empty");

	FormalRegExpEmpty* empty = new FormalRegExpEmpty();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "empty");
	return empty;
}

} /* namespace regexp */

