/*
 * FormalRegExpConcatenation.cpp
 *
 *  Created on: Nov 27, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExpConcatenation.h"
#include "../../exception/AlibException.h"
#include "../unbounded/UnboundedRegExpConcatenation.h"

namespace regexp {

FormalRegExpConcatenation::FormalRegExpConcatenation(FormalRegExpElement&& left, FormalRegExpElement&& right) {
	this->left = std::move(left).plunder();
	this->right = std::move(right).plunder();
}

FormalRegExpConcatenation::FormalRegExpConcatenation(const FormalRegExpElement& left, const FormalRegExpElement& right) {
	this->left = left.clone();
	this->right = right.clone();
}

FormalRegExpConcatenation::FormalRegExpConcatenation(const FormalRegExpConcatenation& other) {
	this->left = other.left->clone();
	this->right = other.right->clone();
}

FormalRegExpConcatenation::FormalRegExpConcatenation(FormalRegExpConcatenation&& other) noexcept : left(std::move(other.left)), right(std::move(other.right)) {
	other.left = NULL;
	other.right = NULL;
	this->attachRegExp(NULL);
}

FormalRegExpConcatenation& FormalRegExpConcatenation::operator=(const FormalRegExpConcatenation& other) {
	if(this == &other) {
		return *this;
	}

	*this = FormalRegExpConcatenation(other);
	
	return *this;
}

FormalRegExpConcatenation& FormalRegExpConcatenation::operator=(FormalRegExpConcatenation&& other) {
	std::swap(this->left, other.left);
	std::swap(this->right, other.right);

	this->left->attachRegExp(this->parentRegExp);
	this->right->attachRegExp(this->parentRegExp);
	
	return *this;
}

FormalRegExpConcatenation::~FormalRegExpConcatenation() noexcept {
	delete left;
	delete right;
}

const FormalRegExpElement & FormalRegExpConcatenation::getLeftElement() const {
	return * left;
}

const FormalRegExpElement & FormalRegExpConcatenation::getRightElement() const {
	return * right;
}

FormalRegExpElement & FormalRegExpConcatenation::getLeftElement() {
	return *left;
}

FormalRegExpElement & FormalRegExpConcatenation::getRightElement() {
	return *right;
}

void FormalRegExpConcatenation::setLeftElement(const FormalRegExpElement& element) {
	FormalRegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete left;
	this->left = elem;
}

void FormalRegExpConcatenation::setLeftElement(FormalRegExpElement&& element) {
	FormalRegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete left;
	this->left = elem;
}

void FormalRegExpConcatenation::setRightElement(const FormalRegExpElement& element) {
	FormalRegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete left;
	this->right = elem;
}

void FormalRegExpConcatenation::setRightElement(FormalRegExpElement&& element) {
	FormalRegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete left;
	this->right = elem;
}

FormalRegExpElement* FormalRegExpConcatenation::clone() const {
	return new FormalRegExpConcatenation(*this);
}

UnboundedRegExpElement* FormalRegExpConcatenation::cloneAsUnbounded() const {
	UnboundedRegExpConcatenation* res = new UnboundedRegExpConcatenation();
	
	UnboundedRegExpElement* leftTmp = left->cloneAsUnbounded();
	res->appendElement(std::move(*leftTmp));
	delete leftTmp;
	
	UnboundedRegExpElement* rightTmp = right->cloneAsUnbounded();
	res->appendElement(std::move(*rightTmp));
	delete rightTmp;

	return res;
}

FormalRegExpElement* FormalRegExpConcatenation::plunder() && {
	return new FormalRegExpConcatenation(std::move(*this));
}

int FormalRegExpConcatenation::compare(const FormalRegExpConcatenation& other) const {
	int res = left->compare(*other.left);
	if(res == 0) res = right->compare(*other.right);
	return res;
}

void FormalRegExpConcatenation::operator>>(std::ostream& out) const {
	out << "(FormalRegExpConcatenation";
	out << " " << *left;
	out << " " << *right;
	out << ")";
}

bool FormalRegExpConcatenation::testSymbol( const alphabet::Symbol & symbol ) const {
	if(left->testSymbol(symbol)) return true;
	if(right->testSymbol(symbol)) return true;
	return false;
}

bool FormalRegExpConcatenation::attachRegExp(const FormalRegExp * regexp ) {
	if(this->parentRegExp == regexp) return true;
	
	this->parentRegExp = regexp;
	if(!left->attachRegExp(regexp)) return false;
	if(!right->attachRegExp(regexp)) return false;
	return true;
}

void FormalRegExpConcatenation::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	left->computeMinimalAlphabet(alphabet);
	right->computeMinimalAlphabet(alphabet);
}

FormalRegExpConcatenation::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace regexp */
