/*
 * StringFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "StringFromXMLParser.h"
#include "../sax/ParserException.h"

#include "../XmlApi.hpp"

namespace string {

String StringFromXMLParser::parseString(std::list<sax::Token>& input) const {
	return parseString(input, std::set<FEATURES>({FEATURES::LINEAR, FEATURES::CYCLIC, FEATURES::EPSILON}));
}

String StringFromXMLParser::parseString(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(alib::xmlApi<Epsilon>::first(input)) {
		if(!features.count(FEATURES::EPSILON)) throw exception::AlibException();
		return String(parseEpsilon(input));
	} else if(alib::xmlApi<LinearString>::first(input)) {
		if(!features.count(FEATURES::LINEAR)) throw exception::AlibException();
		return String(parseLinearString(input));
	} else if(alib::xmlApi<CyclicString>::first(input)) {
		if(!features.count(FEATURES::CYCLIC)) throw exception::AlibException();
		return String(parseCyclicString(input));
	} else {
		throw sax::ParserException(sax::Token("Epsilon, LinearString, CyclicString", sax::Token::TokenType::START_ELEMENT), input.front());
	}
}

bool StringFromXMLParser::first(const std::list<sax::Token>& input) const {
	if(alib::xmlApi<Epsilon>::first(input) || alib::xmlApi<LinearString>::first(input) || alib::xmlApi<CyclicString>::first(input)) {
		return true;
	} else {
		return false;
	}
}

LinearString StringFromXMLParser::parseLinearString(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::STRING_LINEAR_STRING);

	LinearString string;
	std::set<alphabet::Symbol> alphabet = parseAlphabet(input);
	string.setAlphabet(alphabet);

	std::vector<alphabet::Symbol> data = parseContentData(input);
	string.setContent(data);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::STRING_LINEAR_STRING);
	return string;
}

CyclicString StringFromXMLParser::parseCyclicString(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::STRING_CYCLIC_STRING);

	CyclicString string;
	std::set<alphabet::Symbol> alphabet = parseAlphabet(input);
	string.setAlphabet(alphabet);

	std::vector<alphabet::Symbol> data = parseContentData(input);
	string.setContent(data);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::STRING_CYCLIC_STRING);
	return string;
}

Epsilon StringFromXMLParser::parseEpsilon(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::STRING_EPSILON);

	Epsilon epsilon;
	std::set<alphabet::Symbol> alphabet = parseAlphabet(input);
	epsilon.setAlphabet(alphabet);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::STRING_EPSILON);
	return epsilon;
}

std::set<alphabet::Symbol> StringFromXMLParser::parseAlphabet(std::list<sax::Token> &input) const {
	std::set<alphabet::Symbol> alphabet;
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alphabet");
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		alphabet.insert(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
	popToken(input, sax::Token::TokenType::END_ELEMENT, "alphabet");
	return alphabet;
}

std::vector<alphabet::Symbol> StringFromXMLParser::parseContentData(std::list<sax::Token>& input) const {
	std::vector<alphabet::Symbol> data;
	popToken(input, sax::Token::TokenType::START_ELEMENT, "content");
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		data.push_back(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
	popToken(input, sax::Token::TokenType::END_ELEMENT, "content");
	return data;
}

} /* namespace string */
