#include <list>
#include "AutomatonTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "automaton/FSM/DFA.h"
#include "automaton/FSM/ExtendedNFA.h"
#include "automaton/PDA/SinglePopDPDA.h"
#include "automaton/PDA/DPDA.h"

#include "automaton/AutomatonException.h"

#include "factory/XmlDataFactory.hpp"
#include "factory/StringDataFactory.hpp"

#include "alphabet/LabeledSymbol.h"

#include "regexp/RegExp.h"
#include "regexp/unbounded/UnboundedRegExp.h"
#include "regexp/unbounded/UnboundedRegExpElements.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_REGISTRATION( AutomatonTest );

void AutomatonTest::setUp() {
}

void AutomatonTest::tearDown() {
}

void AutomatonTest::testDFAParser() {
	automaton::DFA automaton(automaton::State(1));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));
	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2));
	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), automaton::State(1));

	automaton.addFinalState(automaton::State(3));

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::list<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::list<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::DFA automaton2 = alib::XmlDataFactory::fromTokens<automaton::DFA>(tokens2);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::FSMStringParserTest() {
	{
		std::string input = 	"ENFA a b c d #E\n"
					">0 3|4 5 1|3|4 - 2\n"
					"1 2 - - - -\n"
					"2 3 - - - -\n"
					"3 - - 4 - -\n"
					"4 - 5 - - 5\n"
					"<5 - - - - 3\n";
		automaton::Automaton automaton = alib::StringDataFactory::fromString<automaton::Automaton>(input);

		std::string output = alib::StringDataFactory::toString(automaton);

		CPPUNIT_ASSERT( input == output );

		automaton::Automaton automaton2 = alib::StringDataFactory::fromString<automaton::Automaton>(output);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
	{
		std::string input = 	"MISNFA a b c d\n"
					">0 3|4 5 1|3|4 -\n"
					"1 2 - - -\n"
					"2 3 - - -\n"
					">3 - - 4 -\n"
					"4 - 5 - -\n"
					"<5 - - - 3\n";
		automaton::Automaton automaton = alib::StringDataFactory::fromString<automaton::Automaton>(input);

		std::string output = alib::StringDataFactory::toString(automaton);

		CPPUNIT_ASSERT( input == output );

		automaton::Automaton automaton2 = alib::StringDataFactory::fromString<automaton::Automaton>(output);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
	{
		std::string input = 	"NFA a b c d\n"
					">0 3|4 5 1|3|4 -\n"
					"1 2 - - -\n"
					"2 3 - - -\n"
					"3 - - 4 -\n"
					"4 - 5 - -\n"
					"<5 - - - 3\n";
		automaton::Automaton automaton = alib::StringDataFactory::fromString<automaton::Automaton>(input);

		std::string output = alib::StringDataFactory::toString(automaton);

		CPPUNIT_ASSERT( input == output );

		automaton::Automaton automaton2 = alib::StringDataFactory::fromString<automaton::Automaton>(output);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
	{
		std::string input = 	"DFA a b c d\n"
					">0 3 5 1 -\n"
					"1 2 - - -\n"
					"2 3 - - -\n"
					"3 - - 4 -\n"
					"4 - 5 - -\n"
					"<5 - - - 3\n";
		automaton::Automaton automaton = alib::StringDataFactory::fromString<automaton::Automaton>(input);

		std::string output = alib::StringDataFactory::toString(automaton);

		CPPUNIT_ASSERT( input == output );

		automaton::Automaton automaton2 = alib::StringDataFactory::fromString<automaton::Automaton>(output);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::SinglePopDPDATransitions() {
	automaton::SinglePopDPDA automaton(automaton::State(1), alphabet::symbolFrom("S"));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addStackSymbol(alphabet::symbolFrom("S"));
	automaton.addStackSymbol(alphabet::symbolFrom("X"));
	automaton.addStackSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(automaton::State(3), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(3), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("Y"), automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), /* eps, */ alphabet::symbolFrom("Y"), automaton::State(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(2), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addFinalState(automaton::State(3));

}

void AutomatonTest::DPDATransitions() {
	automaton::DPDA automaton(automaton::State(1), alphabet::symbolFrom("S"));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));
	automaton.addState(automaton::State(4));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addStackSymbol(alphabet::symbolFrom("S"));
	automaton.addStackSymbol(alphabet::symbolFrom("X"));
	automaton.addStackSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(automaton::State(3), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(3), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("Y") }, automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), /* eps, */ { alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(2), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("S") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } ));
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } ),
		exception::AlibException
	);
	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } ),
		exception::AlibException
	);

	automaton.addFinalState(automaton::State(3));
}

void AutomatonTest::testExtendedNFAAlphabet() {
	automaton::ExtendedNFA automaton (automaton::State(0));
	automaton::State s0(0);
	automaton::State s1(1);
	automaton.addState(s0);
	automaton.addState(s1);

	automaton.setInputAlphabet({alphabet::symbolFrom('a'), alphabet::symbolFrom('b') });

	regexp::UnboundedRegExpConcatenation goodConcat1, goodConcat2;
	regexp::UnboundedRegExpConcatenation badConcat;

	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('a')));
	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('d')));

	CPPUNIT_ASSERT_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(badConcat)}, s1), automaton::AutomatonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat1)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat2)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpEpsilon())}, s1));
}

void AutomatonTest::testNPDATransitions() {
	automaton::NPDA automaton(automaton::State(0), alphabet::symbolFrom("T"));

	automaton.setStates({automaton::State(0), automaton::State(1), automaton::State(2), automaton::State(3), automaton::State(4)});
	automaton.setInputAlphabet({alphabet::symbolFrom("a2"), alphabet::symbolFrom("a1"), alphabet::symbolFrom("a0")});
	automaton.setStackAlphabet({alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});

	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(0), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(0), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(0), {});

	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("R"), alphabet::symbolFrom("T")});

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, automaton::State(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, automaton::State(1), {alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(1), {});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, automaton::State(2), {});

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("R")});

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, automaton::State(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, automaton::State(3), {alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(3), {});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, automaton::State(4), {});

	automaton.setFinalStates({automaton::State(4)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testRHPDATransitions() {
	automaton::RealTimeHeightDeterministicNPDA automaton(alphabet::Symbol(alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK));
	automaton.setStates({automaton::State(1), automaton::State(2), automaton::State(3)});
	automaton.setInitialStates({automaton::State(1)});

	automaton.setInputAlphabet({alphabet::symbolFrom("a"), alphabet::symbolFrom("b")});
	automaton.setStackAlphabet({alphabet::Symbol(alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK), alphabet::symbolFrom("X"), alphabet::symbolFrom("Y")});

	automaton.addCallTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2), alphabet::symbolFrom("X"));
	automaton.addCallTransition(automaton::State(2), automaton::State(3), alphabet::symbolFrom("X"));
	automaton.addReturnTransition(automaton::State(3), alphabet::symbolFrom("Y"), automaton::State(1));

	automaton.setFinalStates({automaton::State(3)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}
